import datetime
import glob
import gzip
import logging
import os.path
import os
import pickle
import requests
import shutil
import time
import xml.etree.ElementTree as ET

NS = {'': 'http://scap.nist.gov/schema/feed/vulnerability/2.0',
      'vuln': 'http://scap.nist.gov/schema/vulnerability/0.4'}

logger = logging.getLogger(__name__)


class NVD(object):
    """Access to the National Vulnerability Database.

    https://nvd.nist.gov/

    """

    mirror = 'http://static.nvd.nist.gov/feeds/xml/cve/'
    file = 'nvdcve-2.0-{}.xml.gz'

    earliest = datetime.datetime.today().year - 5  # last five years
    updates = 'Modified'

    def __init__(self, mirror=None, cache_dir=None):
        self.cves = {}
        if mirror:
            self.mirror = mirror
        if not cache_dir:
            cache_dir = '~/.cache/vulnix'

        self.source = self.mirror + self.file
        self.download_path = os.path.expanduser(cache_dir) + '/'

    def update(self):
        if not os.path.exists(self.download_path):
            os.makedirs(self.download_path)
        current_year = datetime.datetime.today().year
        years = list(range(self.earliest, current_year + 1)) + [self.updates]
        for year in years:
            target = self.download_path + '{}.xml'.format(year)
            # don't download it again if archive exists
            if os.path.exists(target + '.cached'):
                # XXX except for current year and modified
                # current update regime by NIST is about eight days
                if year == current_year and \
                        self._isolder(target + '.cached', 8):
                    pass
                elif year == self.updates and \
                        self._isolder(target + '.cached', 8):
                    pass
                else:
                    logger.info('{} is up-to-date.'.format(year))
                    continue
            logger.info('Updating {}'.format(year))
            r = requests.get(self.source.format(year))
            r.raise_for_status()
            with open(target + '.gz', 'wb') as fd:
                for chunk in r.iter_content(1024**2):
                    fd.write(chunk)

    def _isolder(self, file, days):
        if os.path.getmtime(file) < time.time() - days * 24 * 3600:
            return True
        else:
            return False

    def _clean(self):
        for file in glob.glob(self.download_path + '*.xml'):
            os.remove(file)
            os.remove(file + '.gz')

    def _unpack(self):
        for archive in glob.glob(self.download_path + '*.xml.gz'):
            with gzip.open(archive, 'rb') as f_in:
                    with open(archive.strip('.gz'), 'wb') as f_out:
                        shutil.copyfileobj(f_in, f_out)

    def parse(self):
        self._unpack()

        for source in glob.glob(self.download_path + '*.xml'):
            # looking for non-cached vulnerability objects
            if os.path.exists(source):
                self.parse_file(source)
        for cached in glob.glob(self.download_path + '*.cached'):
            with open(cached, 'rb') as fobj:
                for vx in pickle.load(fobj):
                    self.cves[vx.cve_id] = vx

        self._clean()

    def parse_file(self, filename):
        cached_vx = []
        logger.debug("Parsing {}".format(filename))
        tree = ET.parse(filename)
        root = tree.getroot()
        for node in root:
            vx = Vulnerability.from_node(node)
            self.cves[vx.cve_id] = vx
            cached_vx.append(vx)
        # cache results
        with open(filename + '.cached', 'wb') as fobj:
            pickle.dump(cached_vx, fobj)

    def __iter__(self):
        return iter(self.cves.values())


class Vulnerability(object):

    cve_id = None
    affected_products = ()

    def __init__(self):
        self.affected_products = []

    @property
    def url(self):
        return ('https://web.nvd.nist.gov/view/vuln/detail?vulnId={}'.
                format(self.cve_id))

    @staticmethod
    def from_node(node):
        self = Vulnerability()
        self.cve_id = node.get('id')
        for product in node.findall('.//vuln:product', NS):
            cpe = CPE.from_uri(product.text)
            if cpe.product is None:
                # This usually indicates a combination where some vulnerability
                # only applies for a specific operating system *vendor*
                continue
            self.affected_products.append(cpe)
        return self


class CPE(object):

    part = None
    vendor = None
    product = None
    version = None
    update = None
    edition = None
    lang = None

    @staticmethod
    def from_uri(uri):
        self = CPE()
        self.uri = uri
        protocol, identifier = uri.split(':/')
        assert protocol == 'cpe'
        component_list = identifier.split(':')
        components = ['part', 'vendor', 'product', 'version', 'update',
                      'edition', 'lang']
        while component_list:
            component_name = components.pop(0)
            component_value = component_list.pop(0)
            setattr(self, component_name, component_value)
        return self

    def __repr__(self):
        return '<CPE %s>' % self.uri



import tempfile
import logging
import os
import subprocess

_log = logging.getLogger(__name__)


def call(cmd):
    """Call `cmd` and swallow stderr iff returncode is 0."""
    with tempfile.TemporaryFile(prefix='stderr') as capture:
        try:
            output = subprocess.check_output(cmd, stderr=capture)
        except subprocess.CalledProcessError:
            capture.seek(0)
            os.fdopen(2, mode='wb').write(capture.read())
            raise
    return output.decode()


class Store(object):

    def __init__(self):
        self.derivations = {}
        self.product_candidates = {}

    def add_gc_roots(self):
        """Add derivations found for all live GC roots.

        Note that this usually includes old system versions.
        """
        _log.debug('loading all live derivations')
        for d in call(['nix-store', '--gc', '--print-live']).splitlines():
            self.update(d)

    def add_path(self, path):
        """Add the closure of all derivations referenced by a store path."""
        _log.debug('loading derivations referenced by "%s"', path)
        deriver = call(['nix-store', '-qd', path]).strip()
        _log.debug('deriver: %s', deriver)
        if not deriver or deriver == 'unknown-deriver':
            raise RuntimeError(
                'Cannot determine deriver. Is this really a path into the '
                'nix store?', path)
        for candidate in call(['nix-store', '-qR', deriver]).splitlines():
            self.update(candidate)

    def update(self, drv_path):
        if not drv_path.endswith('.drv'):
            return
        if drv_path in self.derivations:
            return
        _log.debug('loading %s', drv_path)
        with open(drv_path) as f:
            d_obj = eval(f.read())
        d_obj.store_path = drv_path
        self.derivations[drv_path] = d_obj
        for product_candidate in d_obj.product_candidates:
            refs = self.product_candidates.setdefault(
                product_candidate, [])
            refs.append(d_obj)


class Derive(object):

    store_path = None

    # This __init__ is compatible with the structure in the derivation file.
    # The derivation files are just accidentally Python-syntax, but hey!
    def __init__(self, output, inputDrvs, inputSrcs, system, builder,  # noqa
                 args, envVars, derivations=None):
        self.output = output
        self.inputDrvs = inputDrvs
        self.inputSrcs = inputSrcs
        self.system = system
        self.builder = builder
        self.args = args
        self.envVars = dict(envVars)
        self.name = self.envVars['name']

        self.affected_by = set()
        self.status = None

    @property
    def is_affected(self):
        return bool(self.affected_by)

    @property
    def simple_name(self):
        # XXX This is a simplification of splitting up the derivation name
        # into its original <name>-<version> form. It's probably going to
        # fail at some point, but we'll cross that bridge then.
        return self.name.rsplit('-', 1)[0]

    @property
    def product_candidates(self):
        variation = self.name.split('-')
        while variation:
            yield '-'.join(variation)
            variation.pop()

    def check(self, vuln, whitelist):
        if vuln in self.affected_by:
            return
        for affected_product in vuln.affected_products:
            if not self.matches(vuln.cve_id, affected_product):
                continue
            if (vuln, affected_product, self) in whitelist:
                continue
            self.affected_by.add(vuln)
            break

    def matches(self, cve_id, cpe):
        # Step 1: determine product name
        prefix = cpe.product + '-'
        if self.name == cpe.product:
            version = None
        elif self.name.startswith(prefix):
            version = self.name.replace(prefix, '', 1)
            if version != cpe.version:
                return False
        else:
            # This product doesn't match at all.
            return False

        # We matched the product and think the version is affected.
        return True

    def roots(self):
        return call(
            ['nix-store', '--query', '--roots', self.store_path]).split('\n')

    def referrers(self):
        return call(['nix-store', '--query', '--referrers',
                     self.store_path]).split('\n')



"""Usage: vulnix {--system | PATH [...]}

vulnix is a tool that scan the NixOS store for packages with known
security issues. There are three main modes of operation:


* Is my NixOS system installation affected?

Invoke:  vulnix --system


* Is my project affected?

Invoke after nix-build:  vulnix ./result


See vulnix --help for a full list of options.
"""

from .nix import Store
from .nvd import NVD
from .whitelist import WhiteList
import click
import logging
import pkg_resources
import subprocess
import sys
import time

DEFAULT_WHITELIST = pkg_resources.resource_filename(
    __name__, 'default_whitelist.yaml')

_log = logging.getLogger(__name__)


class Timer:
    def __enter__(self):
        self.start = time.clock()
        return self

    def __exit__(self, *args):
        self.end = time.clock()
        self.interval = self.end - self.start


def howto():
    head, tail = __doc__.split('\n', 1)
    click.secho(head, fg='yellow')
    click.echo(tail, nl=False)


def output(affected_derivations, verbosity):
    status = []
    derivations = []
    seen = {}
    # we need name-version only once
    for item in affected_derivations:
        marker = item.name
        if marker in seen:
            continue
        seen[marker] = 1
        derivations.append(item)
    derivations.sort(key=lambda k: k.simple_name)

    amount = len(derivations)
    names = ', '.join(derivations[k].simple_name for k in range(3))
    summary = 'Found {} advisories for {}'.format(amount, names)
    if amount > 3:
        summary += ', ... (and {:d} more)'.format(amount - 3)
    click.secho(summary, fg='red')

    for derivation in derivations:
        click.echo('\n{}\n{}\n'.format('=' * 72, derivation.name))
        if verbosity >= 1:
            click.echo(derivation.store_path)
            if verbosity >= 2:
                click.echo()
                click.echo("Referenced by:")
                for referrer in derivation.referrers():
                    click.echo("\t" + referrer)
                click.echo("Used by:")
                for root in derivation.roots():
                    click.echo("\t" + root)
        click.echo("CVEs:")
        for cve in derivation.affected_by:
            click.echo("\t" + cve.url)
        click.echo('=' * 72)
        status.append(1 if derivation.status == 'inprogress' else 2)

    return max(status)


def populate_store(gc_roots, system, paths):
    """Load derivations from nix store depending on cmdline invocation."""
    store = Store()
    if gc_roots:
        store.add_gc_roots()
    if system:
        store.add_path('/nix/var/nix/gcroots/current-system')
        store.add_path('/nix/var/nix/gcroots/booted-system')
    if paths:
        for path in paths:
            store.add_path(path)
    return store


@click.command('vulnix')
@click.option('-G', '--gc-roots', is_flag=True,
              help='Scan all active GC roots (including old ones)')
@click.option('-S', '--system', is_flag=True,
              help='Scan both the booted system and the current system')
@click.option('-w', '--whitelist', multiple=True, type=click.File(),
              help='Add another whiltelist YAML file to declare exceptions.')
@click.option('-m', '--mirror',
              help='Use another mirror for downloading the nvd files. '
              'Default: {}'.format(NVD.mirror))
@click.option('--default-whitelist/--no-default-whitelist', default=True,
              help='Load built-in base whitelist from "{}". Additional '
              'whitelist files can be specified using the "-w" option. '
              'Default: yes'.format(DEFAULT_WHITELIST))
@click.option('-d', '--debug', is_flag=True,
              help='Show debug information.')
@click.option('-v', '--verbose', count=True,
              help='Increase output verbosity.')
@click.option('-c', '--cache-dir', default='~/.cache/vulnix',
              help='Alter the default cache directory')
@click.argument('path', nargs=-1,
                type=click.Path(exists=True))
def main(debug, verbose, whitelist, default_whitelist,
         gc_roots, system, path, mirror, cache_dir):
    """Scans nix store paths for derivations with security vulnerabilities."""
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.getLogger('requests').setLevel(logging.ERROR)
        if verbose:
            logging.basicConfig(level=logging.INFO)
        else:
            logging.basicConfig(level=logging.WARNING)

    if not (gc_roots or system or path):
        howto()
        sys.exit(3)

    _log.debug('loading derivations')
    with Timer() as t:
        try:
            store = populate_store(gc_roots, system, path)
        except (subprocess.CalledProcessError, RuntimeError) as e:
            _log.exception(e)
    _log.debug('store load time: %f', t.interval)

    _log.debug('loading NVD data')
    with Timer() as t:
        if mirror:
            nvd = NVD(mirror=mirror, cache_dir=cache_dir)
        else:
            nvd = NVD(cache_dir=cache_dir)
        nvd.update()
        nvd.parse()
    _log.debug('NVD load time: %f', t.interval)

    wl = WhiteList()
    if default_whitelist:
        with open(DEFAULT_WHITELIST) as f:
            wl.parse(f)
    for fobj in whitelist:
        wl.parse(fobj)

    _log.debug('scanning')
    affected = set()
    with Timer() as t:
        for vuln in nvd:
            for prod in vuln.affected_products:
                for derivation in store.product_candidates.get(
                        prod.product, []):
                    derivation.check(vuln, wl)
                    if derivation.is_affected:
                        affected.add(derivation)
    _log.debug('scan time: %f', t.interval)

    if affected:
        # sensu maps following return codes
        # 0 - ok, 1 - warning, 2 - critical, 3 - unknown
        returncode = output(affected, verbose)
        _log.debug('returncode %d', returncode)
        sys.exit(returncode)
    else:
        click.secho('vulnix: no vulnerabilities detected', fg='green')
